BAOBZI_CASE(1, 6, 0)
BAOBZI_CASE(1, 8, 0)
BAOBZI_CASE(1, 10, 0)
BAOBZI_CASE(1, 12, 0)
BAOBZI_CASE(1, 14, 0)
BAOBZI_CASE(1, 16, 0)
BAOBZI_CASE(2, 6, 0)
BAOBZI_CASE(2, 8, 0)
BAOBZI_CASE(2, 10, 0)
BAOBZI_CASE(2, 12, 0)
BAOBZI_CASE(2, 14, 0)
BAOBZI_CASE(2, 16, 0)
BAOBZI_CASE(3, 6, 0)
BAOBZI_CASE(3, 8, 0)
BAOBZI_CASE(3, 10, 0)
BAOBZI_CASE(3, 12, 0)
BAOBZI_CASE(3, 14, 0)
BAOBZI_CASE(3, 16, 0)
#ifdef BAOBZI_CPU_DISPATCH
BAOBZI_CASE(1, 6, 1)
BAOBZI_CASE(1, 8, 1)
BAOBZI_CASE(1, 10, 1)
BAOBZI_CASE(1, 12, 1)
BAOBZI_CASE(1, 14, 1)
BAOBZI_CASE(1, 16, 1)
BAOBZI_CASE(2, 6, 1)
BAOBZI_CASE(2, 8, 1)
BAOBZI_CASE(2, 10, 1)
BAOBZI_CASE(2, 12, 1)
BAOBZI_CASE(2, 14, 1)
BAOBZI_CASE(2, 16, 1)
BAOBZI_CASE(3, 6, 1)
BAOBZI_CASE(3, 8, 1)
BAOBZI_CASE(3, 10, 1)
BAOBZI_CASE(3, 12, 1)
BAOBZI_CASE(3, 14, 1)
BAOBZI_CASE(3, 16, 1)
#endif
#ifdef BAOBZI_CPU_DISPATCH
BAOBZI_CASE(1, 6, 2)
BAOBZI_CASE(1, 8, 2)
BAOBZI_CASE(1, 10, 2)
BAOBZI_CASE(1, 12, 2)
BAOBZI_CASE(1, 14, 2)
BAOBZI_CASE(1, 16, 2)
BAOBZI_CASE(2, 6, 2)
BAOBZI_CASE(2, 8, 2)
BAOBZI_CASE(2, 10, 2)
BAOBZI_CASE(2, 12, 2)
BAOBZI_CASE(2, 14, 2)
BAOBZI_CASE(2, 16, 2)
BAOBZI_CASE(3, 6, 2)
BAOBZI_CASE(3, 8, 2)
BAOBZI_CASE(3, 10, 2)
BAOBZI_CASE(3, 12, 2)
BAOBZI_CASE(3, 14, 2)
BAOBZI_CASE(3, 16, 2)
#endif
#ifdef BAOBZI_CPU_DISPATCH
BAOBZI_CASE(1, 6, 3)
BAOBZI_CASE(1, 8, 3)
BAOBZI_CASE(1, 10, 3)
BAOBZI_CASE(1, 12, 3)
BAOBZI_CASE(1, 14, 3)
BAOBZI_CASE(1, 16, 3)
BAOBZI_CASE(2, 6, 3)
BAOBZI_CASE(2, 8, 3)
BAOBZI_CASE(2, 10, 3)
BAOBZI_CASE(2, 12, 3)
BAOBZI_CASE(2, 14, 3)
BAOBZI_CASE(2, 16, 3)
BAOBZI_CASE(3, 6, 3)
BAOBZI_CASE(3, 8, 3)
BAOBZI_CASE(3, 10, 3)
BAOBZI_CASE(3, 12, 3)
BAOBZI_CASE(3, 14, 3)
BAOBZI_CASE(3, 16, 3)
#endif
